
#include <time.h>

#include "clock.h"
#include "xgstypes.h"
#include "xgsstate.h"
#include "emul.h"
#include "video.h"

// CLOCK.CPP
byte	clk_data_reg;
byte	clk_ctl_reg;
byte	bram[256];
tCLKCURRTIME clk_curr_time;

int	clk_state;
int	clk_addr;

#define CLOCK_DEBUG(x)

int CLK_init()
{
   SETFUNCTION("CLK_init");
	FILE     *fp;
   char     *path=NULL;
   int      rv = 0;

   WriteDebug("Starting Clock Emulation");

   path = ConvertBSTR(g_bstrBRAMFile);
   if (path)
   {
      sprintf(emul_buffer,"Attempting To Load BRAM File '%s'......",path);
      WriteDebug(emul_buffer);

	   fp = fopen(path,"rb");
	   if (fp)
      {
		   if (fread(bram,1,256,fp) == 256) 
         {
            WriteDebug("Done");
            rv = 0;
         };
         fclose(fp);
      }
      else
      {
         DoErrorMessage("Unable to Load BRAM File. Please check your configuration.");
         WriteDebug("FAILED!");
      };
      free(path);
      path=NULL;
   }
   else
   {
      DoErrorMessage("Unable to Load BRAM file. No filename was configured. Please check your configuration.");
      WriteDebug("No BRAM File Specified...");
   };

   return(rv);
};

void CLK_update()
{
}

void CLK_reset()
{
	clk_data_reg = 0;
	clk_ctl_reg &= 0x3F;
	clk_state = 0;
}

void CLK_shutdown()
{
   SETFUNCTION("CLK_shutdown");
	FILE	*fp;
	char	*path=NULL;
   WriteDebug("Shutting down clock emulation");
	path = ConvertBSTR(g_bstrBRAMFile);
   if (path)
   {
      sprintf(emul_buffer,"Saving BRAM Settings to '%s'.....",path);
      WriteDebug(emul_buffer);

	   fp = fopen(path,"wb");
      if (fp)
      {
         fwrite(bram,1,256,fp);
         fclose(fp);
         WriteDebug("Done");
      }
      else
         WriteDebug("Unable To Open Output File");
      free(path);
      path = NULL;
   }
   else
      WriteDebug("No BRAM File Specified To Save");
};

byte CLK_getData(byte val)
{
	return clk_data_reg;
}

byte CLK_setData(byte val)
{
	clk_data_reg = val;
	return 0;
}

byte CLK_getControl(byte val)
{
	byte ret_val;
	ret_val=(clk_ctl_reg | VID_getBorder(0));
	return(ret_val);
}

byte CLK_setControl(byte val)
{
		VID_setBorder((byte) (val & 0x0F));
	clk_ctl_reg = val & 0xF0;

	/* need to make sure there is a data transfer going on if not then just
	   zero the data register. */
	if (!(clk_ctl_reg & 0x20))
		clk_data_reg = 0;

	switch (clk_state & 0x0F) 
   {
		case 0 : 
         
         if ((clk_data_reg & 0x78) == 0x38) 
         {
					clk_state = (clk_data_reg & 0x80) | 0x41;
					clk_addr = (clk_data_reg & 0x07) << 5;
			} 
         else 
         if ((clk_data_reg & 0x73) == 0x01) 
         {
					clk_state = (clk_data_reg & 0x80) | 0x02;
					clk_addr = (clk_data_reg >> 2) & 0x03;
			}
			break;
	
   	case 1 :

				clk_addr |= ((clk_data_reg >> 2) & 0x1F);
				clk_state++;
				break;

		case 2 : 
         
            if (clk_state & 0x40) 
            {
					if (clk_state & 0x80) 
               {
						clk_data_reg = bram[clk_addr];
					}
               else 
               {
						bram[clk_addr] = clk_data_reg;
					}
				} 
            else 
            {
					if (clk_state & 0x80) 
               {
						if (!clk_addr) 
                     clk_curr_time.L = time(NULL) + CLK_OFFSET;
//						clk_data_reg = clk_curr_time.B[3-clk_addr];
						clk_data_reg = clk_curr_time.B[clk_addr];
					} 
               else 
               {
						/* Can't set system time */
					}
				}
				clk_state = 0;
				break;
		default :	
            clk_state = 0;
				break;
	}
	clk_ctl_reg &= 0x7F;	/* Clear transaction bit since we're done */
	return 0;

}
