
#include "xgsdispatch.h"
#include "xgstypes.h"
#include "xgsstate.h"
#include "cpu.h"
#include "cpumicro.h"
#include "xgsmemory.h"
#include "emul.h"
#include "debugger.h"
#include "vid-drv.h"

// DISPATCH.CPP
duala		atmp,opaddr;
dualw		wtmp,otmp,operand;
int		a1,a2,a3,a4,o1,o2,o3,o4;
void		(**cpu_curr_opcode_table)();

long g_lDispatchDebug=0;
long g_lDebugRun=0;

void CPU_run(void)
{

	static   last_update,next_update;
	int	   opcode;

   // If we are "unpausing" the emulator, then we need to skip over the reset code and
   // go straight to the opcode dispatch section
   if (g_lEmulatorSignal==EMUL_DORESUME)
   {
      g_lEmulatorSignal = EMUL_RUNNING;
      goto dispatch;
   };

   // reset the cpu information.
	cpu_cycle_count = 0;
	last_update = 0;
	next_update = cpu_update_period;
	E = 1;
	F_setM(1);
	F_setX(1);
	CPU_modeSwitch();

dispatch:

   // if we have a signal, then exit the dispatcher
   if (g_lEmulatorSignal)
      return;

   // if we have executed "x" number of cycles, then update the hardware
	if (cpu_cycle_count >= next_update) goto update;

   if (g_lDebugMode==2)
   {
      
      if ((g_lDebugHalt) && (!g_lDebugRun))
      {
         DEBUGGER_Update();
         while((g_lDebugHalt) && (!g_lEmulatorSignal) && (g_lDebugMode==2))
         {
            MSG msg;
            if (PeekMessage(&msg,g_hVidHWND,0,0,PM_REMOVE))
            {
               VIDWindowProc(msg.hwnd,msg.message,msg.wParam,msg.lParam);
            };
         
            UpdateWindow(g_hVidHWND);
         }
         g_lDebugHalt=1;    
      }
      else
      {
         DEBUGGER_Update();
      };
  
   };

update_resume:

debug_resume:

   // lets just assume that this call never occurs... stopping here will halt the emulator!
	//if (cpu_stop) goto dispatch;

	if (cpu_reset) goto reset;
	if (cpu_abort) goto abort;
	if (cpu_nmi) goto nmi;
	if (cpu_irq) goto irq;

irq_return:

   // lets assume that this code doesn't get executed. We should never halt or pause the CPU
   // execution because the entire emulator core is dependent on it!
	//if (cpu_wait) goto dispatch;

   // read our next opcode
	opcode = M_READ(PC.A);
	PC.W.PC++;

   // execute the opcode
	(**cpu_curr_opcode_table[opcode])();

	goto dispatch;

/* Special cases. Since these don't happen a lot more often than they	*/
/* do happen, accessing them this way means most of the time the	*/
/* generated code is _not_ branching. Only during the special cases do	*/
/* we take the branch penalty (if there is one).			*/

update:

   // update our hardware
	E_UPDATE(cpu_cycle_count - last_update);
	last_update = cpu_cycle_count;
	next_update = last_update + cpu_update_period;
	goto update_resume;

debug:
	CPU_debug();
	goto debug_resume;

reset:
	(**cpu_curr_opcode_table[256])();
	goto dispatch;

abort:
	(**cpu_curr_opcode_table[257])();
	goto dispatch;

nmi:
	(**cpu_curr_opcode_table[258])();
	goto dispatch;

irq:
	if (P & 0x04) goto irq_return;
	(**cpu_curr_opcode_table[259])();
	goto dispatch;

}

/* Recalculate opcode_offset based on the new processor mode */

void CPU_modeSwitch(void) {

	int	opcode_offset;

	if (E) {
		opcode_offset = 1040;
	} else {
		if (F_getX) {
			X.B.H = 0;
			Y.B.H = 0;
		}
		opcode_offset = ((~P >> 4) & 0x03) * 260;
	}
#ifdef OLDCYCLES
	cpu_curr_cycle_table = cpu_cycle_table + opcode_offset;
#endif
	cpu_curr_opcode_table = (void (__cdecl ** )(void))cpu_opcode_table + opcode_offset;
}
