
#include "emul.h"
#include "xgstypes.h"
#include "xgsstate.h"
#include "cpu.h"
#include "xgsmemory.h"
#include "video.h"
#include "clock.h"
#include "sound.h"
#include "video.h"
#include "smtport.h"
#include "iwm.h"
#include "adb.h"
#include "vid-drv.h"
#include "adb-drv.h"
#include "snd-drv.h"


// emul state
int	   emul_period;
int	   emul_vbl;
int	   emul_vblirq;
int	   emul_qtrsecirq;
int	   emul_onesecirq;
int	   emul_scanirq;
int	   emul_speed;
int	   emul_speed2;
int	   emul_delay;
int	   emul_tick;
int	   emul_microtick;
int	   emul_vbl_count;
char	   emul_buffer[256];
int	   emul_target_cycles; 
double	emul_target_speed;
double	emul_times[60];
double	emul_total_time;
int	   emul_cycles[60];
word32	emul_last_cycles;
word32	emul_total_cycles;
word32	emul_last_time;
word32	emul_this_time;
word32   emul_this_cycles;

DWORD    g_lRunThreadID;


//UINT EMUL_EmulatorRunThread( LPVOID pParam )
DWORD WINAPI EMUL_EmulatorRunThread(LPVOID lpParameter)
{
	// start the emulator.... I have encapsulated the call inside of this function
	// so we can implement a seperate worker thread later if necessary...
   //

restartafterpause:

   EMUL_run();

   ////////////////////////////////////////////////////////////////////////////
   // when EMUL_EmulatorRunThread() exits, there are two possible states of the variable
   // g_lEmulatorSignal..
   //
   // EMUL_DOQUIT:   Simply Quit the Emulator... Simple
   // EMUL_DOPAUSE:  Pause the emulator. Pause the video, sound and input.
   ////////////////////////////////////////////////////////////////////////////

   // if we are pausing then we need to keep everything alive except for the video and
   // input stuff.  
   if (g_lEmulatorSignal==EMUL_DOPAUSE)
   {

      g_lEmulatorSignal=EMUL_PAUSED;

      ADB_PauseInput();
      SND_PauseSound();
      VID_PauseVideo();

      // while we are paused, sleep 100ms.
      while (g_lEmulatorSignal==EMUL_PAUSED)
         Sleep(100);

      if (g_lEmulatorSignal==EMUL_DORESUME)
      {
         VID_RestoreVideoAfterPause();
         SND_RestoreSoundAfterPause();
         ADB_RestoreInputAfterPause();
         goto restartafterpause;
      };

      // if we get this far, then we must have been paused, and the user chose to close
      // the window... in this case, we want to reset the emulator state to EMUL_IDLE
      g_lEmulatorSignal=EMUL_IDLE;      

   }
   else
   {
      // we are shutting down the emulator!
      EMUL_Shutdown();
   };

   g_lRunThreadID = 0;

   return 1;
}

long EMUL_getCurrentTime()
{
	// return the tick count since startup...
	// personally, I don't think this function is necessary, since this code
	// is no longer portable anyway...
	return (long) GetTickCount();
}

void EMUL_CalcNativeSpeed()
{
   SETFUNCTION("EMUL_CalcNativeSpeed");
	long		lCycleCount=0;
	long		lStartTick;
	long		lEndTick;

   // Perform a simple loop to determine how fast the computer is. We will execute the loop for no less
   // than 1024ms and increment a counter to determine how many times we actually do it..
   WriteDebug("Performing CPU Speed Test......");
	lEndTick=0;
	lStartTick=GetTickCount();
	while((lEndTick-lStartTick) < 1024)
	{
		lCycleCount++;
		lEndTick=GetTickCount();
	};
   WriteDebug("Done")
	g_lNativeCPUSpeed=lCycleCount;
	sprintf(emul_buffer,"NativeCPUSpeed=%d (higher=faster)",g_lNativeCPUSpeed);
	WriteDebug(emul_buffer);
};

void EMUL_Initialize()
{
   SETFUNCTION("EMUL_Initialize");
   WriteDebug("Initializing Emulator Core");
	STATE_Initialize();
};

void EMUL_Shutdown()
{
   SETFUNCTION("EMUL_Shutdown");
   WriteDebug("Shutting down the emulator");
	SMPT_shutdown();
	IWM_shutdown();
	CLK_shutdown();
	SND_shutdown();
	MEM_shutdown();
	ADB_shutdown();
	VID_shutdown();
	STATE_Shutdown();
   g_lEmulatorSignal=EMUL_IDLE;
}

void EMUL_doVBL()
{
	emul_vbl_count++;

	if (emul_vbl == 0x00) 
   {
		if (emul_vblirq) 
      {
			if (!(mem_diagtype & 0x08)) 
         {
				mem_diagtype |= 0x08;
				CPU_addIRQ();
			}
		}

		VID_update();
		MEM_update();
		ADB_update();
		CLK_update();
		IWM_update();
		SMPT_update();
	}

	if ((emul_tick == 0) || (emul_tick == 15) || (emul_tick == 30) || (emul_tick == 45)) 
   {
		if (emul_qtrsecirq) 
      {
			if (!(mem_diagtype & 0x10)) {
				mem_diagtype |= 0x10;
				CPU_addIRQ();
			}
		}
	}

	if (emul_onesecirq && !emul_tick) 
   {
		if (!(vid_vgcint & 0x40)) {
			vid_vgcint |= 0xC0;
			CPU_addIRQ();
		}
	}

	if (++emul_tick == 60) emul_tick = 0;

}

void WND_update()
{
   MSG      msg;
   while (PeekMessage(&msg,g_hVidHWND,0,0,PM_NOREMOVE))
   {
      if (GetMessage(&msg,g_hVidHWND,0,0) != -1)
      {
         TranslateMessage(&msg); 
         DispatchMessage(&msg);
      };
   };
};


void EMUL_hardwareUpdate(word32 cycles)
{
   static int     counter=0;
   int            new_period;
	double		   dDiff;
	word32         lCurrentTime;
   long           lTickCount;
   long           lDelayLength;

   // update our sound registers
	SND_update();

   // update our window
   WND_update();

	emul_microtick++;
	counter++;

	// every 64 hardware updates, we are going to implement a delay to throttle
	// the run speed of the emulator. 
	if (counter==64)
	{

      // find out how many ticks have passed since our last update
		lCurrentTime         = GetTickCount();
		emul_this_time       = lCurrentTime - emul_last_time;
      emul_this_cycles     = (cpu_cycle_count - emul_last_cycles);

      // if our clock looped over, then assume that the cycles is the same as the last update 
      // period
      if (emul_this_time<0)
         emul_this_time=cycles;

      // update our counters...
		emul_total_time      += emul_this_time;
   	emul_total_cycles    += emul_this_cycles;
		emul_last_cycles     = cpu_cycle_count;
		emul_last_time       = lCurrentTime;
		counter=0;

      // calc our average run speed
		g_dAverageRunSpeed   = (emul_total_cycles / emul_total_time)  / 1024;	

      
      // if we are using dynamic period allocation, then throttle the update
      // period. This will cause the CPU to "float" in efficiency, but
      // allow more updates per second during high load conditions
      if (!g_lEmulatorUpdateMode)
      {
         // if we are running fast, then we need to increase the update period so we
         // can keep running fast
	      new_period = g_dAverageRunSpeed * (32.0);
	      if (new_period < 100) new_period = 100;
	      if (new_period > 1000) new_period = 1000;
      }
      else
      {
         new_period = g_lEmulatorUpdatePeriod;
      };

	   emul_period = new_period;
	   CPU_setUpdatePeriod(new_period);

      // if we are throttling our run speed, then we need to implement a delay here
		if (g_dTargetRunSpeed>0)
      {
         // calculate our speed excess and generate a delay amount
		   dDiff = g_dAverageRunSpeed-g_dTargetRunSpeed;
         lDelayLength = (g_lNativeCPUSpeed/1024000)*dDiff*64;
         if (lDelayLength>0)
			{
            lTickCount = lDelayLength;
            while(lTickCount>0)
            {
               lTickCount--;
               sprintf(emul_buffer,"This is a simple delay stub that will decrease the run speed of the emulator by a pre-specified amount%d",lTickCount);
            };
         };
      };
	};


   // handle our microtick so we can update the vbl
   switch (emul_microtick)
   {
      case 400:

		      emul_vbl = 0x80;
		      vid_vert_cnt = emul_microtick >> 1;
            break;

      case 524:

		      emul_microtick = 0;
		      vid_vert_cnt = 0;
		      emul_vbl = 0x00;
		      EMUL_doVBL();
            break;

      default:

      		vid_vert_cnt = emul_microtick >> 1;
            break;
   };

	if (emul_scanirq && (vid_vert_cnt < 200) && (slow_memory[0x019D00 + vid_vert_cnt] & 0x40)) 
   {
		if (!(vid_vgcint & 0x20)) {
			vid_vgcint |= 0xA0;
			CPU_addIRQ();
		}
	}

}


void EMUL_start()
{
   SETFUNCTION("EMUL_start");

   /////////////////////////////////////////////////////////////////////////
   //
   // on this call, the emulator can be in three possible states..
   // (indicated by the value of g_lEmulatorSignal
   // 
   // 0 = idle                      (not currently running)
   // 1 = terminate emulator        (signals the emulator to terminate)
   // 2 = pause emulator            (signals the emulator to exit, but not terminate)
   // 3 = resume emulator operation (signals the emulator to resume operation after a pause)
   //
   /////////////////////////////////////////////////////////////////////////

   // if the emulator is currently paused, then restore out video and 
   // input processes...
   if (g_lEmulatorSignal==EMUL_PAUSED)   
   {
      long lThisTime=GetTickCount();
      long lTimePaused = (lThisTime-emul_last_time);
      emul_last_time=lThisTime;
      // signal the emulator to resume operation after pause
      g_lEmulatorSignal=EMUL_DORESUME;
   }   
   else
   {
      // spawn a seperate thread that will run the emulator...
      CreateThread(NULL,NULL,EMUL_EmulatorRunThread,NULL,NULL,&g_lRunThreadID);
      sprintf(emul_buffer,"Emulator Run Thread Created... id=0x%xd",g_lRunThreadID);
      WriteDebug(emul_buffer);
   }; 

};


void EMUL_pause()
{
   // tell the emulator to pause...
   g_lEmulatorSignal=EMUL_DOPAUSE;
};


void EMUL_stop()
{
   // kill the emulator...
   g_lEmulatorSignal=EMUL_DOQUIT;
   while(g_lEmulatorSignal != EMUL_IDLE)
   {
      Sleep(500);
   };
};


int EMUL_run()
{
   SETFUNCTION("EMUL_run");
	char	*path;
	int	err,i,j,unit;

	WriteDebug("Running the emulator");

   // if the emulator is idle, then start everything up...
   if (g_lEmulatorSignal==EMUL_IDLE)
   {

      EMUL_CalcNativeSpeed();

	   if ((err = MEM_init()))
      {
         DoErrorMessage("Unable to Initialize Memory\n");
         return err;
      };

	   if ((err = VID_init())) 
      {
		   MEM_shutdown();
         DoErrorMessage("Unable to Initialize Video\n");
		   return err;
	   };

	   if ((err = SND_init())) 
      {
		   VID_shutdown();
		   MEM_shutdown();
         DoErrorMessage("Unable to Intialize Sound\n");
		   return err;
	   };

	   if ((err = ADB_init())) 
      {
		   SND_shutdown();
		   VID_shutdown();
		   MEM_shutdown();
         DoErrorMessage("Unable to Initialize ADB\n");
		   return err;
	   };

	   if ((err = CLK_init())) 
      {
		   ADB_shutdown();
		   SND_shutdown();
		   VID_shutdown();
		   MEM_shutdown();
         DoErrorMessage("Unable to Initialize Clock\n");
		   return err;
	   };

	   if ((err = IWM_init())) 
      {
		   CLK_shutdown();
		   ADB_shutdown();
		   SND_shutdown();
		   VID_shutdown();
		   MEM_shutdown();
         DoErrorMessage("Unable to Initialize IWM\n");
		   return err;
	   };

	   if ((err = SMPT_init())) 
      {
		   IWM_shutdown();
		   CLK_shutdown();
		   ADB_shutdown();
		   SND_shutdown();
		   VID_shutdown();
		   MEM_shutdown();
         DoErrorMessage("Unable to Initialize SMPT\n");
		   return err;
	   };

	   emul_last_time = EMUL_getCurrentTime();

	   for (i = 0 ; i < 60 ; i++) 
      {
		   emul_times[i] = 0.0;
		   emul_cycles[i] = 0;
	   };

	   emul_total_time = 0.0;
	   emul_total_cycles = 0;
	   emul_tick = 0;
	   emul_microtick = 0;

	   emul_last_cycles = 0;

	   emul_vbl = 0;
	   emul_vblirq = 0;
	   emul_qtrsecirq = 0;
	   emul_onesecirq = 0;
	   emul_speed = 1;
	   emul_speed2 = 1;

	   emul_target_cycles = 2500000;
	   emul_target_speed = 2.5;

	   EMUL_reset();

	   for (i = 0 ; i < NUM_SMPT_DEVS ; i++) 
      {
		   if (!g_bstrSmartPort[i]) continue;
		   path = ConvertBSTR(g_bstrSmartPort[i]);
		   SMPT_loadDrive(i, path);
	   }

		for (i=5; i < 7; i++)
			for (j=1; j < 3; j++)
			{
				if (!g_bstrSlot[i][j]) continue;
				path = ConvertBSTR(g_bstrSlot[i][j]);
				if (path)
					IWM_loadDrive(i,j,path);				
			};		

	   emul_period = g_lNativeCPUSpeed/142245;
      if (emul_period < 64)
         emul_period=64;
	   CPU_setUpdatePeriod(emul_period);
      sprintf(emul_buffer,"Initial Update Period Is %d cycles",emul_period);
      WriteDebug(emul_buffer);	   
      g_lEmulatorSignal=EMUL_RUNNING;
   };


   WriteDebug("*****EMULATOR IS RUNNING*****")
	CPU_run();
};

void EMUL_reset()
{
	emul_delay = 0;
	MEM_reset();
	VID_reset();
	SND_reset();
	ADB_reset();
	CLK_reset();
	IWM_reset();
	SMPT_reset();
	CPU_reset();
}


void EMUL_trace(int parm)
{
	CPU_setTrace(parm);
}

void EMUL_nmi()
{
	int	i;
	for (i = 0x0000 ; i < 0x0200 ; i++) {
		if (!mem_pages[i].readPtr) continue;
		memcpy(mem_pages[0xE800+i].writePtr,mem_pages[i].readPtr,256);
	}
 	CPU_nmi();
}

void EMUL_handleWDM(byte parm)
{
	switch(parm) {
		case 0xC7 :	SMPT_prodosEntry();
				break;
		case 0xC8 :	SMPT_smartportEntry();
				break;
		case 0xFD :	CPU_setTrace(0);
				break;
		case 0xFE :	CPU_setTrace(1);
				break;
		case 0xFF :	EMUL_Shutdown();
				break;
		default :	break;
	}
}

byte EMUL_getVBL(byte val)
{	
	return emul_vbl;
};

