/*

   COPYRIGHT NOTIFICATION

   This information is copyrighted by
   Steven W. Mentzer

   You may not sell or assume ownership of this information 
   without my express written consent.

   This information can be freely redistributed, published, compiled, used or modified.

   Inquiries should be directed to   steve@staticmonkey.com

*/

#include <ddraw.h>

#include "vid-drv.h"
#include "xgstypes.h"
#include "xgsstate.h"
#include "adb.h"
#include "emul.h"
#include "xgsmemory.h"
#include "video.h"
#include "hires.h"
#include "snd-drv.h"
#include "adb-drv.h"
#include "vid-drv_utils.h"
#include "vid-drv_pixel.h"

//#define VIDDEBUG
long g_lVID_InService= FALSE;  
long g_lUpdateTimerID=0;
long g_lChangeVideoMode=0;

// VID_DRV.CPP
PIXEL	*vid_lines[2048];
vid_color vid_textcolors[16] = 
{
	{ 0x00, 0x00, 0x00 },
	{ 0xDD, 0x00, 0x33 },
	{ 0x00, 0x00, 0x99 },
	{ 0xDD, 0x22, 0xDD },
	{ 0x00, 0x77, 0x22 },
	{ 0x55, 0x55, 0x55 },
	{ 0x22, 0x22, 0xFF },
	{ 0x66, 0xAA, 0xFF },
	{ 0x88, 0x55, 0x00 },
	{ 0xFF, 0x66, 0x00 },
	{ 0xAA, 0xAA, 0xAA },
	{ 0xFF, 0x99, 0x88 },
	{ 0x00, 0xDD, 0x00 },
	{ 0xFF, 0xFF, 0x00 },
	{ 0x55, 0xFF, 0x99 },
	{ 0xFF, 0xFF, 0xFF }
};
vid_color vid_supercolors[256];

LPDIRECTDRAW         g_pDD;
LPDIRECTDRAW2        g_pDD2;
LPDIRECTDRAWSURFACE  g_pDDBackBuffer;
LPDIRECTDRAWSURFACE  g_pDDPrimarySurface;
PALETTEENTRY			g_pDDPaletteEntry[256];
LPDIRECTDRAWPALETTE	g_pDDPalette;
DDSURFACEDESC        g_DDBackBufferSurfaceDesc;
DDSURFACEDESC        g_DDPrimarySurfaceDesc;

tRGB16DEF   g_RGB16DEF;

int	vid_out_width,vid_out_height;
int   vid_out_x,vid_out_y;

long  g_lVidDebug=0;

BOOL VID_LockService(BOOL bWait)
{
   long lWaitCounter=0;
   if (bWait==TRUE)
   {
      while ((InterlockedExchange (&g_lVID_InService, TRUE) == FALSE) && (lWaitCounter<1024000))
         lWaitCounter++;
      return g_lVID_InService;
   }
   else
   {
      if (InterlockedExchange (&g_lVID_InService, TRUE) == FALSE)
         return FALSE;
      else
         return TRUE;
   };
   return FALSE;
};

BOOL VID_UnlockService()
{
   g_lVID_InService=FALSE;
   return TRUE;
};


void VID_InvalidateVideoMemory()
{
   // tell the code to redraw the entire screen next time...
   memset(mem_slowram_changed,0xFF,sizeof(mem_slowram_changed));
};


long FAR PASCAL VIDWindowProc(HWND hWnd,UINT message,WPARAM wParam,LPARAM lParam)
{
   int i, done;
   LPWAVEHDR pWaveHdr;
   SETFUNCTION("VIDWindowProc");

   switch(message)
   {

      case MM_WOM_DONE:

         pWaveHdr = (LPWAVEHDR) lParam;
	      if (pWaveHdr->dwFlags == (WHDR_DONE | WHDR_PREPARED)) 
         {
		      i = 0;
            done = 0;
		      while ((i < g_lSoundOutputBuffers) && (!done)) 
            {
			      if (pWaveHdr == mybuffers[i].lpWaveHdr) 
               {
				      mybuffers[i].busy = FALSE;
				      done = 1;
			      }
			      i++;
		      }	
         };

         break;
		case WM_CHAR:
			ADB_inputKeyDown((int) wParam, 0);
			break;

		case WM_KEYDOWN:
			ADB_inputKeyDown((int) wParam, 1);
			break;

		case WM_KEYUP:
			ADB_inputKeyRelease((int) wParam, 1);
			break;

		case WM_LBUTTONDOWN:
			ADB_inputLeftMouseDown();
			break;

		case WM_LBUTTONUP:
			ADB_inputLeftMouseUp();
			break;

		case WM_RBUTTONDOWN:
			ADB_inputRightMouseDown();
			break;

		case WM_RBUTTONUP:
			ADB_inputRightMouseUp();
			break;

		case WM_MOUSEMOVE:
			ADB_inputMotionNotify(LOWORD(lParam), HIWORD(lParam));
			break;

      case WM_DESTROY:
            break;
   };
   return DefWindowProc(hWnd, message, wParam, lParam);

} 

long g_lParentWindow=0;
void VID_SetParentWindow(long lParentHwnd)
{
   g_lParentWindow = lParentHwnd;
};


char *GetErrorMessage(DWORD errnum)
{
   LPVOID lpMsgBuf;
   char *pszOut=NULL;
   FormatMessage( 
       FORMAT_MESSAGE_ALLOCATE_BUFFER | 
       FORMAT_MESSAGE_FROM_SYSTEM | 
       FORMAT_MESSAGE_IGNORE_INSERTS,
       NULL,
       errnum,
       MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
       (LPTSTR) &lpMsgBuf,
       0,
       NULL);

   pszOut = (char*)malloc(strlen((LPCTSTR)lpMsgBuf)+1);
   if (pszOut)
   {
      strcpy(pszOut,(LPCTSTR)lpMsgBuf);
   };
   LocalFree(lpMsgBuf);
   return pszOut;
};
 


int VID_outputInit(void)
{
   SETFUNCTION("VID_outputInit");
  
   WNDCLASS		      wc;
   HRESULT           hr;
   char *errmsg=NULL;

	WriteDebug("Initializing Windows Video Driver");

   // register an instance of a window class
   wc.style = CS_HREDRAW | CS_VREDRAW;
   wc.lpfnWndProc = VIDWindowProc;
   wc.cbClsExtra = 0;
   wc.cbWndExtra = 0;
   wc.hInstance = g_hDLLInstance;
   wc.hIcon = NULL;
   wc.hCursor = NULL;
   wc.hbrBackground = NULL;
   wc.lpszMenuName = NULL;
   wc.lpszClassName = WINDOW_NAME;
   hr = RegisterClass( &wc );

   // create a window using the Win32 API and get the handle
   if (!g_lParentWindow)
      g_hVidHWND = CreateWindowEx(0,WINDOW_NAME,WINDOW_TITLE,WS_OVERLAPPEDWINDOW,CW_USEDEFAULT,CW_USEDEFAULT,128,128,NULL,NULL,g_hDLLInstance,NULL);
   else
      g_hVidHWND = CreateWindowEx(0,WINDOW_NAME,WINDOW_TITLE,WS_POPUP,CW_USEDEFAULT,CW_USEDEFAULT,128,128,(HWND)g_lParentWindow,NULL,g_hDLLInstance,NULL);

   // if the handle is NULL, then the window was not created...
   if( !g_hVidHWND )
   {
      DoErrorMessage("Unable to Create A Child Window",1);         
      WriteDebug("Unable To Create Client Window");
      return FALSE;
   };

   // if we are in windowed mode, then just show the window. If we are fullscreen, maximize it.
   if (g_lVidWindowed)
   {
      HDC dc = GetDC(g_hVidHWND);
      if (dc)
      {
         g_lVidBitDepth = GetDeviceCaps(dc,BITSPIXEL);
         sprintf(emul_buffer,"Windows Desktop Bit Depth is %dbpp",g_lVidBitDepth);
         WriteDebug(emul_buffer);
         ReleaseDC(g_hVidHWND,dc);
      };
      ShowWindow(g_hVidHWND,SW_SHOW);
   }
   else
      ShowWindow(g_hVidHWND,SW_SHOWMAXIMIZED);

   // Create a directdraw component...
	DDCHECKCOM(DirectDrawCreate(NULL, &g_pDD, NULL),"Error Creating DirectDraw Interface");

   // Get the DirectDraw2 interface on the directdraw root object
	DDCHECKCOM(g_pDD->QueryInterface(IID_IDirectDraw2,(void**) &g_pDD2),"Error Obtaining DirectDraw2 interface on DirectDraw root object");

   // if we are windowed, then our cooperative level is NORMAL, otherwise, we are exclusive fullscreen
   if (g_lVidWindowed)
	   DDCHECKCOM(g_pDD2->SetCooperativeLevel(g_hVidHWND, DDSCL_NORMAL),"Unable to set windowed cooperative level of NORMAL")
    else
      DDCHECKCOM(g_pDD2->SetCooperativeLevel(g_hVidHWND, DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN | DDSCL_ALLOWREBOOT),"Unable to set DirectDraw fullscreen cooperative level of EXCLUSIVE");

   // if we are fullscreen, set our display mode...
   if (!g_lVidWindowed)
   {
      sprintf(emul_buffer,"Setting Display Mode To %d*%d*%d @ %dHz",g_lVidWidth,g_lVidHeight,g_lVidBitDepth,g_lVidRefreshRate);
	   WriteDebug(emul_buffer);
      DDCHECKCOM(g_pDD2->SetDisplayMode(g_lVidWidth,g_lVidHeight,g_lVidBitDepth,g_lVidRefreshRate,NULL),"Unable to set DirectDraw display mode");
   };

   // create our primary surface...
	g_DDPrimarySurfaceDesc.dwSize = sizeof(g_DDPrimarySurfaceDesc);
	g_DDPrimarySurfaceDesc.dwFlags = DDSD_CAPS;
	g_DDPrimarySurfaceDesc.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;
	DDCHECKCOM(g_pDD2->CreateSurface(&g_DDPrimarySurfaceDesc, &g_pDDPrimarySurface, NULL),"Unable to create our primary surface");

   // create a backbuffer surface, so we can blt our video from it to our primary surface later...
   DDCHECKCOM(DDCreateOffscreenSurface(g_pDD,g_lVidWidth,g_lVidHeight,&g_pDDBackBuffer,&g_DDBackBufferSurfaceDesc),"Failed to create the backbuffer surface");   
 
   // if we are running in 8 bit color mode (god help you), then create a palette object.
   // I know full well, that this will not work properly in windowed mode, since palettes
   // must be shared to work correctly. I am not going to write code to support 8 bit
   // desktops.
   if (g_lVidBitDepth==8)
	{
      WriteDebug("Video Mode is using 8 bit color depth, Creating Palette Object");
	   DDCHECKCOM(g_pDD2->CreatePalette(DDPCAPS_8BIT, g_pDDPaletteEntry, &g_pDDPalette, NULL),"Unable to create fullscreen 8 bit palette");
      DDCHECKCOM(g_pDDPrimarySurface->SetPalette(g_pDDPalette),"Unable to set palette");
   };

   // if we are running in 16 bit color mode, then we need to determine how the 16 bit mask works
   // some boards use 5-6-5, some 5-5-5. I makes a difference....
   if(g_lVidBitDepth==16)
   {
      WriteDebug("Video Mode is using 16 bit color depth. Analyzing structure, and formatting color bit shift operators");
      GetRGB16DEF(g_pDDBackBuffer);
   };


   // each possible bit depth has a corresponding pixel-plot routine.
   // assign our function array pointers to the correct routines.
   switch (g_lVidBitDepth)
   {

      case 32:
         VID_PixelPlot[VID_PIXEL_SUPER]=VID_PP_32_SUPER;
         VID_PixelPlot[VID_PIXEL_BORDER]=VID_PP_32_BORDER;
         VID_PixelPlot[VID_PIXEL_TEXT]=VID_PP_32_TEXT;
         VID_PixelPlot[VID_PIXEL_HIRES]=VID_PP_32_HIRES;
         VID_PixelPlot[VID_PIXEL_DBL]=VID_PP_32_DBL;
         break;
      case 24:
         VID_PixelPlot[VID_PIXEL_SUPER]=VID_PP_24_SUPER;
         VID_PixelPlot[VID_PIXEL_BORDER]=VID_PP_24_BORDER;
         VID_PixelPlot[VID_PIXEL_TEXT]=VID_PP_24_TEXT;
         VID_PixelPlot[VID_PIXEL_HIRES]=VID_PP_24_HIRES;
         VID_PixelPlot[VID_PIXEL_DBL]=VID_PP_24_DBL;
         break;
      case 16:
         VID_PixelPlot[VID_PIXEL_SUPER]=VID_PP_16_SUPER;
         VID_PixelPlot[VID_PIXEL_BORDER]=VID_PP_16_BORDER;
         VID_PixelPlot[VID_PIXEL_TEXT]=VID_PP_16_TEXT;
         VID_PixelPlot[VID_PIXEL_HIRES]=VID_PP_16_HIRES;
         VID_PixelPlot[VID_PIXEL_DBL]=VID_PP_16_DBL;
         break;
      case 8:
         VID_PixelPlot[VID_PIXEL_SUPER]=VID_PP_8_SUPER;
         VID_PixelPlot[VID_PIXEL_BORDER]=VID_PP_8_BORDER;
         VID_PixelPlot[VID_PIXEL_TEXT]=VID_PP_8_TEXT;
         VID_PixelPlot[VID_PIXEL_HIRES]=VID_PP_8_HIRES;
         VID_PixelPlot[VID_PIXEL_DBL]=VID_PP_8_DBL;
         break;
      default:
         WriteDebug("Unknown Bit Depth! Unable to initialize video pixel plot functions");
         return 1;
         break;
   };


   // if we are windowed, resize it appropriately....
   // one thing that needs to be done is the calculation of the menu bar and frame sizes.
   // we need a true client area of 640x480, and if we set a window to be 640x480, the actual client size
   // will be smaller because of the frame.
   if (g_lVidWindowed)
   {
      WriteDebug("Setting Windowed Mode");

      long lFrameWidth=0;
      long lFrameHeight=0;
      long lMenuHeight=0;
      RECT  rWindow;
      RECT  rClient;

      if (g_lParentWindow)
      {
         SetParent(g_hVidHWND,(HWND)g_lParentWindow);
      };

      SetWindowPos(g_hVidHWND,0,0,0,g_lVidWidth,g_lVidHeight,0);
      
      GetWindowRect(g_hVidHWND,&rWindow);
      GetClientRect(g_hVidHWND,&rClient);

      lFrameWidth = (g_lVidWidth - rClient.right)/2;
      lFrameHeight = lFrameWidth;
      lMenuHeight = (g_lVidHeight - rClient.bottom - lFrameHeight);
      
      if(!SetWindowPos(g_hVidHWND,0,0,0,g_lVidWidth+(lFrameWidth*2),g_lVidHeight+(lFrameHeight+lMenuHeight),0))
         WriteDebug("Unable to set window position");

      LPDIRECTDRAWCLIPPER lpddc;
      
      hr = g_pDD2->CreateClipper(NULL,&lpddc,NULL);
      if FAILED(hr)
         WriteDebug("unable to create clipper");
      hr = lpddc->SetHWnd(NULL,g_hVidHWND);
      if FAILED(hr) 
         WriteDebug("unable to set hwnd");
      
      hr = g_pDDPrimarySurface->SetClipper(lpddc);     
      if FAILED(hr) 
         WriteDebug("unable to set clipper");
   };

   UpdateWindow(g_hVidHWND);

   // if we aren't directly writing to video memory via the emulator run thread,
   // then spawn a seperate multimedia timer (thread) to draw graphic info.
   if (!g_lLockVideoUpdates)
   {

      // create a timer that updates the screen periodically...
      long lTimerInterval = (1000/((int)g_dTargetVideoRefreshRate));
      g_lUpdateTimerID = timeSetEvent(lTimerInterval,lTimerInterval,VID_UpdateProc,NULL,TIME_PERIODIC);

      if (!g_lUpdateTimerID)
      {
         WriteDebug("Unable to create multimedia timer for screen refresh!");
         return 1;
      }
      else
      {
         sprintf(emul_buffer,"Multimedia Timer ID#0x%Xd Created With Interval of %dms",g_lUpdateTimerID,lTimerInterval);
         WriteDebug(emul_buffer);
      };
   }
   else
      WriteDebug("Video Updates are Fixed on CPU Cycles");

   
   WriteDebug("Successful");
	return 0;

}

void VID_PauseVideo()
{
   VID_outputShutdown();
};

void VID_RestoreVideoAfterPause()
{
   VID_outputInit();
   VID_outputResize(vid_out_width,vid_out_height);
};


// VID_UpdateProc() is called by a timer set up to execute this code /n/ times per second.
//
// It does the following
//
//  1 - Checks for re-entrancy. If the last repaint didn't finish, dont do it now...
//  2 - 
// 

void CALLBACK VID_UpdateProc(UINT uID, UINT uMsg, DWORD dwUser,DWORD dw1, DWORD dw2)
{
   SETFUNCTION("VID_updateProc");
   HRESULT                 hr;
   static long             lFPSCounter=0;
   static long             lStartTime=GetTickCount();
   LPDIRECTDRAWSURFACE     pDDSurface;
   DDSURFACEDESC           *pDDSurfaceDesc;
   static long             lLastBLT;
   static long             lLastLock;

   // Check to see if we are ok to draw. If g_lVID_InService is true, then we are not done with the previous paint.
   if (VID_LockService(FALSE)==TRUE)
   {     

      // if the emulator is actually running...
      if ((g_lEmulatorSignal==EMUL_RUNNING) && (!g_lChangeVideoMode))
      {

         VID_GetDrawingSurface(&pDDSurface,&pDDSurfaceDesc);

         // increase our FPS counter. If our counter equals the target number of frames per second, then 
         // calculate the number of milliseconds it took, and generate a true FPS rate.
         lFPSCounter++;
         if (
            (lFPSCounter==((int)g_dTargetVideoRefreshRate)) 
            || 
            ((g_lLockVideoUpdates) && (lFPSCounter==30))
            )
         {
            long lElapsed;
            double dSeconds;
            lElapsed=GetTickCount()-lStartTime;
            dSeconds = (((double)lElapsed)/1000);
            if (dSeconds>0)
               g_dAverageVideoRefreshRate=lFPSCounter/ dSeconds;
            lFPSCounter=0;
            lStartTime=GetTickCount();
         };

         // Lock our backbuffer....
         // if we are running in fullscreen mode, our backbuffer is really our
         // screen surface. this is a little trick I play to improve performance.
	      hr = pDDSurface->Lock(NULL, pDDSurfaceDesc, DDLOCK_WAIT, NULL);
	      if (hr != DD_OK) 
	      {
		      // if we lost the surface, reacquire it...
		      if (hr == DDERR_SURFACELOST) 
		      {
			      WriteDebug("Lost the Drawing surface...reaquiring");
			      hr = pDDSurface->Restore();
			      if (hr != DD_OK) return;
			      VID_outputResize(vid_out_width, vid_out_height);
		      } 
		      else 
		      {
			      WRITEERRORSTRING(hr,"VID_UpdateProc(): Couldn't Lock Drawing Surface");
               WriteDDError(hr);
               g_lVID_InService=FALSE;
               return;
		      }
	      }

	      vid_xmin = VID_WIDTH;
	      vid_ymin = VID_HEIGHT;
	      vid_xmax = vid_ymax = 0;

         // Call the video update routine for this screen mode...
         // this call will "paint" each updated pixel to our backbuffer
	      if (VID_updateRoutine)
            (*VID_updateRoutine)();

         // if we are showing diagnostic information via f7, we need to paint it as well
         if (g_lShowStatus)
            VID_outputStatus1(NULL);

         DDCHECKCOMNR(pDDSurface->Unlock(pDDSurfaceDesc->lpSurface),"Couldn't unlock drawing Surface");

         // if we are running in windowed mode, we must then blt our backbuffer to our child window
         // todo: properly use clipper objects...
         if (g_lVidWindowed)
         {  
               RECT  destRect;
               RECT  srcRect;
               POINT pt;
               GetClientRect( g_hVidHWND, &destRect );
               pt.x = pt.y = 0;
               srcRect.bottom = g_lVidHeight;
               srcRect.top = 0;
               srcRect.left = 0;
               srcRect.right= g_lVidWidth;
               ClientToScreen( g_hVidHWND, &pt );
               OffsetRect(&destRect, pt.x, pt.y);
               hr = g_pDDPrimarySurface->Blt(&destRect,g_pDDBackBuffer,&srcRect,NULL,NULL);
               if FAILED(hr)
               {
                  WriteDebug("Error during Windowed BLT: ");
                  WriteDDError(hr);
               };
         };
         

      };

      VID_UnlockService();
   }  
   else
   {
      //WriteDebug("VID_UpdateProc(): Reentrant call!");
   }   

};


void VID_outputImage()
{
   SETFUNCTION("VID_outputImage");

   // this function is mostly useless now, since the timer thread handles it all...
   if (g_lChangeVideoMode)
   {
      switch(g_lChangeVideoMode)
      {

         case 1: // switch from fullscreen to windowed, vice-versa
                  
                  WriteDebug("Pausing Video");
                  VID_PauseVideo();
                  SND_PauseSound();
                  ADB_PauseInput();
                  g_lChangeVideoMode=0;
                  g_lVidWindowed = !g_lVidWindowed;
                  WriteDebug("Restoring Video");
                  VID_RestoreVideoAfterPause();
                  SND_RestoreSoundAfterPause();
                  ADB_RestoreInputAfterPause();                  
                  VID_InvalidateVideoMemory();
                  break;
         default:
                  break;
      };
   };

   if (g_lLockVideoUpdates)
      VID_UpdateProc(0,0,0,0,0);

};

void VID_outputShutdown()
{
   SETFUNCTION("VID_outputShutdown");
   HRESULT  hr;

   
   if (g_lUpdateTimerID)
   {
      hr = timeKillEvent(g_lUpdateTimerID);
      if (hr != TIMERR_NOERROR)
      {
         sprintf(emul_buffer,"Unable to kill Multimedia timer ID#%Xd",g_lUpdateTimerID);
         WriteDebug(emul_buffer);
      };
      g_lUpdateTimerID=0;
   };


	if (g_pDDPalette)
   {
      WriteDebug("Shutting Down DirectDrawPalette");
      g_pDDPalette->Release();
      g_pDDPalette=NULL;
   };
   if (g_pDDBackBuffer)
   {
      WriteDebug("Shutting Down DirectDraw Back Buffer");
	   g_pDDBackBuffer->Release();
      g_pDDBackBuffer=NULL;
   }

   if(g_pDDPrimarySurface)
   {
      WriteDebug("Shutting Down Direct Draw Primary Surface");
	   g_pDDPrimarySurface->Release();
      g_pDDPrimarySurface=NULL;
   };

   if (g_pDD2)
   {
      WriteDebug("Shutting Down DirectDraw2 Root Object");
	   g_pDD2->RestoreDisplayMode();
	   g_pDD2->Release();
      g_pDD2=NULL;
   };
   if (g_pDD)
   {
      WriteDebug("Shutting Down DirectDraw root Object");
	   g_pDD->Release();
      g_pDD=NULL;
   };

   if (g_hVidHWND)
   {
      WriteDebug("Destroying Child Window\n");
      DestroyWindow(g_hVidHWND);
      g_hVidHWND = NULL;
   };

   g_lVID_InService=FALSE;

}

void VID_outputStatus1(char *msg)
{
   HDC dc;
   HRESULT hr;
   if (g_lShowStatus)
   {
      LPDIRECTDRAWSURFACE  pDDSurface;
      DDSURFACEDESC        *pDDSurfaceDesc;
      VID_GetDrawingSurface(&pDDSurface,&pDDSurfaceDesc);
      sprintf(emul_buffer,"pdl0=%d  pdl1=%d   UpdatePeriod=%d   VidLockMode=%d   AverageFPS=%2.1f  AverageRunSpeed=%2.1fMhz\0",adb_pdl0, adb_pdl1, emul_period,g_lLockVideoUpdates,g_dAverageVideoRefreshRate,g_dAverageRunSpeed);
      hr = pDDSurface->GetDC(&dc);
      if SUCCEEDED(hr)
      {
         TextOut(dc,3,g_lVidHeight-20,emul_buffer,strlen(emul_buffer));
         pDDSurface->ReleaseDC(dc);
      }; 
   };
}

void VID_outputStatus2(char *msg)
{
}

HRESULT VID_DrawBorder()
{
   SETFUNCTION("VID_DrawBorder");
   HRESULT hr;
	int		y,x;
   long     lColorIndex=252; // 252 is the palette entry for the border
   LPDIRECTDRAWSURFACE  pDDSurface;
   DDSURFACEDESC        *pDDSurfaceDesc;

   if (VID_LockService(TRUE)==TRUE)
   {
   

      VID_GetDrawingSurface(&pDDSurface,&pDDSurfaceDesc);

      // lock the display
	   DDCHECKCOMNR(pDDSurface->Lock(NULL, pDDSurfaceDesc, DDLOCK_WAIT, NULL),"VID_DrawBorder(): Could Not Lock The BackBuffer");

      if (SUCCEEDED(hr))
      {
         // generate a vid_lines array representing the entire screen...   
         VID_GenerateVidLines(2);

         // Draw the top border...
         for (y=0; y < vid_out_y; y++)
         { 
            x = 0;
            while(x<g_lVidWidth)
            {
               VID_putPixel(x,y,lColorIndex,VID_PIXEL_BORDER);
               x++;
            };
         };

         // left and right borders
	      for (y = vid_out_y ; y < vid_out_y + vid_out_height; y++) 
         {
            x=0;
            while(x < vid_out_x)
            { VID_putPixel(x,y,lColorIndex,VID_PIXEL_BORDER); x++; };
            x = vid_out_x+vid_out_width;
            while(x < g_lVidWidth)
            { VID_putPixel(x,y,lColorIndex,VID_PIXEL_BORDER); x++; };
         };

         // draw the bottom border
	      for (y = vid_out_y + vid_out_height ; y < g_lVidHeight; y++) 
         {
            x=0;
            while (x<g_lVidWidth)
            { VID_putPixel(x,y,lColorIndex,VID_PIXEL_BORDER); x++; };
         };

         // reset our vidlines for our regular viewport.
         VID_GenerateVidLines(1);

	      DDCHECKCOMNR(pDDSurface->Unlock(pDDSurfaceDesc->lpSurface),"VID_DrawBorder(): Could not unlock the BackBuffer");

         VID_InvalidateVideoMemory();

      };

      VID_UnlockService();

   }
   else
      return E_FAIL;

   return S_OK;
};

void VID_outputResize(int width, int height)
{

   // calculate the border widths and center position of our screen area...
	vid_out_width = width;
	vid_out_height = height;
	vid_out_x = (g_lVidWidth - width) / 2;
	vid_out_y = (g_lVidHeight - height) / 2;

   // redraw our new border...
   VID_DrawBorder();
   
}


void VID_outputSetTextColors()
{
	g_pDDPaletteEntry[253].peRed = vid_textcolors[vid_textbgcolor].red;
	g_pDDPaletteEntry[253].peGreen = vid_textcolors[vid_textbgcolor].green;
	g_pDDPaletteEntry[253].peBlue = vid_textcolors[vid_textbgcolor].blue;
	
	g_pDDPaletteEntry[254].peRed = vid_textcolors[vid_textfgcolor].red;
	g_pDDPaletteEntry[254].peGreen = vid_textcolors[vid_textfgcolor].green;
	g_pDDPaletteEntry[254].peBlue = vid_textcolors[vid_textfgcolor].blue;

	if (g_lVidBitDepth==8)
   {
      g_pDDPalette->SetEntries(0, 0, 256, g_pDDPaletteEntry);
   }
   else
   {
      VID_outputResize(vid_out_width, vid_out_height);
   };

}

void VID_outputSetBorderColor()
{
	g_pDDPaletteEntry[252].peRed = vid_textcolors[vid_bordercolor].red;
	g_pDDPaletteEntry[252].peGreen = vid_textcolors[vid_bordercolor].green;
	g_pDDPaletteEntry[252].peBlue = vid_textcolors[vid_bordercolor].blue;

	if (g_lVidBitDepth==8)
   {
      g_pDDPalette->SetEntries(0, 0, 256, g_pDDPaletteEntry);
   }
   else
   {
      VID_outputResize(vid_out_width, vid_out_height);
   };
}

void VID_outputSetStandardColors()
{
	int i;

	for (i = 0; i < 16; i++) 
   {
		g_pDDPaletteEntry[i].peRed = vid_textcolors[i].red;
		g_pDDPaletteEntry[i].peGreen = vid_textcolors[i].green;
		g_pDDPaletteEntry[i].peBlue = vid_textcolors[i].blue;
	}
   
	g_pDDPaletteEntry[253].peRed = vid_textcolors[vid_textbgcolor].red;
	g_pDDPaletteEntry[253].peGreen = vid_textcolors[vid_textbgcolor].green;
	g_pDDPaletteEntry[253].peBlue = vid_textcolors[vid_textbgcolor].blue;
	
	g_pDDPaletteEntry[254].peRed = vid_textcolors[vid_textfgcolor].red;
	g_pDDPaletteEntry[254].peGreen = vid_textcolors[vid_textfgcolor].green;
	g_pDDPaletteEntry[254].peBlue = vid_textcolors[vid_textfgcolor].blue;

	g_pDDPaletteEntry[252].peRed = vid_textcolors[vid_bordercolor].red;
	g_pDDPaletteEntry[252].peGreen = vid_textcolors[vid_bordercolor].green;
	g_pDDPaletteEntry[252].peBlue = vid_textcolors[vid_bordercolor].blue;
	
	if (g_lVidBitDepth==8)
   {  
      g_pDDPalette->SetEntries(0, 0, 256, g_pDDPaletteEntry);
   }
   else
   {
      VID_outputResize(vid_out_width, vid_out_height);
   };

	vid_black = 0;
	vid_white = 15;
}

void VID_outputSetSuperColors()
{
	int addr, i;
	int requireupdate = 0;
 
	addr = 0x019E00;
	for (i = 0 ; i < 256 ; i++) 
   {
		if (mem_slowram_changed[addr >> 8] & mem_change_masks[addr & 0xFF]) 
      {
			g_pDDPaletteEntry[i].peBlue = (slow_memory[addr] & 0x0F) * 17;
			g_pDDPaletteEntry[i].peGreen = ((slow_memory[addr] >> 4) & 0x0F) * 17;
			g_pDDPaletteEntry[i].peRed = (slow_memory[addr+1] & 0x0F) * 17;
			requireupdate = 1;
		}
		addr += 2;
	}

   if (requireupdate) 
   { 
      
      if (g_lVidBitDepth==8)
      {
         g_pDDPalette->SetEntries(0, 0, 256, g_pDDPaletteEntry);
      }
      else
      {
         VID_InvalidateVideoMemory();
      };
   };

	mem_slowram_changed[0x019E] = 0;
	mem_slowram_changed[0x019F] = 0;
	
}
